@@ -31,9 +31,12 @@ G_DEFINE_TYPE (MMDevice, mm_device, G_TYPE_OBJECT);
 
 enum {
     PROP_0,
+    PROP_PATH,
     PROP_UDEV_DEVICE,
     PROP_PLUGIN,
     PROP_MODEM,
+    PROP_HOTPLUGGED,
+    PROP_VIRTUAL,
     PROP_LAST
 };
 
@@ -47,9 +50,14 @@ static GParamSpec *properties[PROP_LAST];
 static guint signals[SIGNAL_LAST];
 
 struct _MMDevicePrivate {
+    /* Whether the device is real or virtual */
+    gboolean virtual;
+
+    /* Device path */
+    gchar *path;
+
     /* Parent UDev device */
     GUdevDevice *udev_device;
-    gchar *udev_device_path;
     guint16 vendor;
     guint16 product;
 
@@ -68,6 +76,12 @@ struct _MMDevicePrivate {
 
     /* When exported, a reference to the object manager */
     GDBusObjectManagerServer *object_manager;
+
+    /* Whether the device was hot-plugged. */
+    gboolean hotplugged;
+
+    /* Virtual ports */
+    gchar **virtual_ports;
 };
 
 /*****************************************************************************/
@@ -116,6 +130,7 @@ get_device_ids (GUdevDevice *device,
     GUdevDevice *parent = NULL;
     const gchar *vid = NULL, *pid = NULL, *parent_subsys;
     gboolean success = FALSE;
+    char *pci_vid = NULL, *pci_pid = NULL;
 
     parent = g_udev_device_get_parent (device);
     if (parent) {
@@ -141,8 +156,9 @@ get_device_ids (GUdevDevice *device,
                 success = TRUE;
                 goto out;
             } else if (g_str_has_prefix (parent_subsys, "usb") &&
-                       g_str_equal (g_udev_device_get_driver (parent), "qmi_wwan")) {
-                /* Need to look for vendor/product in the parent of the QMI device */
+                       (!g_strcmp0 (g_udev_device_get_driver (parent), "qmi_wwan") ||
+                        !g_strcmp0 (g_udev_device_get_driver (parent), "cdc_mbim"))) {
+                /* Need to look for vendor/product in the parent of the QMI/MBIM device */
                 GUdevDevice *qmi_parent;
 
                 qmi_parent = g_udev_device_get_parent (parent);
@@ -151,6 +167,19 @@ get_device_ids (GUdevDevice *device,
                     pid = g_udev_device_get_property (qmi_parent, "ID_MODEL_ID");
                     g_object_unref (qmi_parent);
                 }
+            } else if (g_str_equal (parent_subsys, "pci")) {
+                const char *pci_id;
+
+                /* We can't always rely on the model + vendor showing up on
+                 * the PCI device's child, so look at the PCI parent.  PCI_ID
+                 * has the format "1931:000C".
+                 */
+                pci_id = g_udev_device_get_property (parent, "PCI_ID");
+                if (pci_id && strlen (pci_id) == 9 && pci_id[4] == ':') {
+                    vid = pci_vid = g_strdup (pci_id);
+                    pci_vid[4] = '\0';
+                    pid = pci_pid = g_strdup (pci_id + 5);
+                }
             }
         }
     }
@@ -194,19 +223,22 @@ get_device_ids (GUdevDevice *device,
 out:
     if (parent)
         g_object_unref (parent);
+    g_free (pci_vid);
+    g_free (pci_pid);
     return success;
 }
 
-static gchar *
-get_driver_name (GUdevDevice *device)
+const gchar *
+mm_device_utils_get_port_driver (GUdevDevice *udev_port)
 {
-    GUdevDevice *parent = NULL;
     const gchar *driver, *subsys;
-    gchar *ret = NULL;
+    const char *name = g_udev_device_get_name (udev_port);
 
-    driver = g_udev_device_get_driver (device);
+    driver = g_udev_device_get_driver (udev_port);
     if (!driver) {
-        parent = g_udev_device_get_parent (device);
+        GUdevDevice *parent;
+
+        parent = g_udev_device_get_parent (udev_port);
         if (parent)
             driver = g_udev_device_get_driver (parent);
 
@@ -218,32 +250,37 @@ get_driver_name (GUdevDevice *device)
             if (subsys && !strcmp (subsys, "bluetooth"))
                 driver = "bluetooth";
         }
+
+        if (parent)
+            g_object_unref (parent);
     }
 
-    if (driver)
-        ret = g_strdup (driver);
-    if (parent)
-        g_object_unref (parent);
+    /* Newer kernels don't set up the rfcomm port parent in sysfs,
+     * so we must infer it from the device name.
+     */
+    if (!driver && strncmp (name, "rfcomm", 6) == 0)
+        driver = "bluetooth";
 
-    return ret;
+    return driver;
 }
 
 static void
 add_port_driver (MMDevice *self,
                  GUdevDevice *udev_port)
 {
-    gchar *driver = NULL;
+    const gchar *driver;
     guint n_items;
     guint i;
 
-    driver = get_driver_name (udev_port);
+    driver = mm_device_utils_get_port_driver (udev_port);
+    if (!driver)
+        return;
 
     n_items = (self->priv->drivers ? g_strv_length (self->priv->drivers) : 0);
     if (n_items > 0) {
         /* Add driver to our list of drivers, if not already there */
         for (i = 0; self->priv->drivers[i]; i++) {
             if (g_str_equal (self->priv->drivers[i], driver)) {
-                g_free (driver);
                 driver = NULL;
                 break;
             }
@@ -255,7 +292,7 @@ add_port_driver (MMDevice *self,
 
     self->priv->drivers = g_realloc (self->priv->drivers,
                                      (n_items + 2) * sizeof (gchar *));
-    self->priv->drivers[n_items] = driver;
+    self->priv->drivers[n_items] = g_strdup (driver);
     self->priv->drivers[n_items + 1] = NULL;
 }
 
@@ -276,7 +313,7 @@ mm_device_grab_port (MMDevice    *self,
                              &self->priv->vendor,
                              &self->priv->product)) {
             mm_dbg ("(%s) could not get vendor/product ID",
-                    self->priv->udev_device_path);
+                    self->priv->path);
         }
     }
 
@@ -393,7 +430,9 @@ export_modem (MMDevice *self)
                                          G_DBUS_OBJECT_SKELETON (self->priv->modem));
 
     mm_dbg ("Exported modem '%s' at path '%s'",
-            g_udev_device_get_sysfs_path (self->priv->udev_device),
+            (self->priv->virtual ?
+             self->priv->path :
+             g_udev_device_get_sysfs_path (self->priv->udev_device)),
             path);
 
     /* Once connected, dump additional debug info about the modem */
@@ -402,7 +441,9 @@ export_modem (MMDevice *self)
             mm_base_modem_get_plugin (self->priv->modem),
             (mm_base_modem_get_vendor_id (self->priv->modem) & 0xFFFF),
             (mm_base_modem_get_product_id (self->priv->modem) & 0xFFFF),
-            g_udev_device_get_subsystem (self->priv->udev_device));
+            (self->priv->virtual ?
+             "virtual" :
+             g_udev_device_get_subsystem (self->priv->udev_device)));
 
     g_free (path);
 }
@@ -454,17 +495,31 @@ mm_device_create_modem (MMDevice                  *self,
     g_assert (self->priv->modem == NULL);
     g_assert (self->priv->object_manager == NULL);
 
-    if (!self->priv->port_probes) {
-        g_set_error (error,
-                     MM_CORE_ERROR,
-                     MM_CORE_ERROR_FAILED,
-                     "Not creating a device without ports");
-        return FALSE;
-    }
+    if (!self->priv->virtual) {
+        if (!self->priv->port_probes) {
+            g_set_error (error,
+                         MM_CORE_ERROR,
+                         MM_CORE_ERROR_FAILED,
+                         "Not creating a device without ports");
+            return FALSE;
+        }
 
-    mm_info ("Creating modem with plugin '%s' and '%u' ports",
-             mm_plugin_get_name (self->priv->plugin),
-             g_list_length (self->priv->port_probes));
+        mm_info ("Creating modem with plugin '%s' and '%u' ports",
+                 mm_plugin_get_name (self->priv->plugin),
+                 g_list_length (self->priv->port_probes));
+    } else {
+        if (!self->priv->virtual_ports) {
+            g_set_error (error,
+                         MM_CORE_ERROR,
+                         MM_CORE_ERROR_FAILED,
+                         "Not creating a virtual device without ports");
+            return FALSE;
+        }
+
+        mm_info ("Creating virtual modem with plugin '%s' and '%u' ports",
+                 mm_plugin_get_name (self->priv->plugin),
+                 g_strv_length (self->priv->virtual_ports));
+    }
 
     self->priv->modem = mm_plugin_create_modem (self->priv->plugin, self, error);
     if (self->priv->modem) {
@@ -486,7 +541,7 @@ mm_device_create_modem (MMDevice                  *self,
 const gchar *
 mm_device_get_path (MMDevice *self)
 {
-    return self->priv->udev_device_path;
+    return self->priv->path;
 }
 
 const gchar **
@@ -510,12 +565,14 @@ mm_device_get_product (MMDevice *self)
 GUdevDevice *
 mm_device_peek_udev_device (MMDevice *self)
 {
+    g_return_val_if_fail (self->priv->udev_device != NULL, NULL);
     return self->priv->udev_device;
 }
 
 GUdevDevice *
 mm_device_get_udev_device (MMDevice *self)
 {
+    g_return_val_if_fail (self->priv->udev_device != NULL, NULL);
     return G_UDEV_DEVICE (g_object_ref (self->priv->udev_device));
 }
 
@@ -596,13 +653,69 @@ mm_device_get_port_probe_list (MMDevice *self)
     return copy;
 }
 
+gboolean
+mm_device_get_hotplugged (MMDevice *self)
+{
+    return self->priv->hotplugged;
+}
+
+/*****************************************************************************/
+
+void
+mm_device_virtual_grab_ports (MMDevice *self,
+                              const gchar **ports)
+{
+    g_return_if_fail (ports != NULL);
+    g_return_if_fail (self->priv->virtual);
+
+    /* Setup drivers array */
+    self->priv->drivers = g_malloc (2 * sizeof (gchar *));
+    self->priv->drivers[0] = g_strdup ("virtual");
+    self->priv->drivers[1] = NULL;
+
+    /* Keep virtual port names */
+    self->priv->virtual_ports = g_strdupv ((gchar **)ports);
+}
+
+const gchar **
+mm_device_virtual_peek_ports (MMDevice *self)
+{
+    g_return_val_if_fail (self->priv->virtual, NULL);
+
+    return (const gchar **)self->priv->virtual_ports;
+}
+
+gboolean
+mm_device_is_virtual (MMDevice *self)
+{
+    return self->priv->virtual;
+}
+
 /*****************************************************************************/
 
 MMDevice *
-mm_device_new (GUdevDevice *udev_device)
+mm_device_new (GUdevDevice *udev_device,
+               gboolean hotplugged)
 {
+    g_return_val_if_fail (udev_device != NULL, NULL);
+
     return MM_DEVICE (g_object_new (MM_TYPE_DEVICE,
                                     MM_DEVICE_UDEV_DEVICE, udev_device,
+                                    MM_DEVICE_PATH, g_udev_device_get_sysfs_path (udev_device),
+                                    MM_DEVICE_HOTPLUGGED, hotplugged,
+                                    NULL));
+}
+
+MMDevice *
+mm_device_virtual_new (const gchar *path,
+                       gboolean hotplugged)
+{
+    g_return_val_if_fail (path != NULL, NULL);
+
+    return MM_DEVICE (g_object_new (MM_TYPE_DEVICE,
+                                    MM_DEVICE_PATH, path,
+                                    MM_DEVICE_HOTPLUGGED, hotplugged,
+                                    MM_DEVICE_VIRTUAL, TRUE,
                                     NULL));
 }
 
@@ -610,7 +723,7 @@ static void
 mm_device_init (MMDevice *self)
 {
     /* Initialize private data */
-    self->priv = G_TYPE_INSTANCE_GET_PRIVATE ((self),
+    self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
                                               MM_TYPE_DEVICE,
                                               MMDevicePrivate);
 }
@@ -624,10 +737,13 @@ set_property (GObject *object,
     MMDevice *self = MM_DEVICE (object);
 
     switch (prop_id) {
+    case PROP_PATH:
+        /* construct only */
+        self->priv->path = g_value_dup_string (value);
+        break;
     case PROP_UDEV_DEVICE:
         /* construct only */
         self->priv->udev_device = g_value_dup_object (value);
-        self->priv->udev_device_path = g_strdup (g_udev_device_get_sysfs_path (self->priv->udev_device));
         break;
     case PROP_PLUGIN:
         g_clear_object (&(self->priv->plugin));
@@ -637,6 +753,12 @@ set_property (GObject *object,
         g_clear_object (&(self->priv->modem));
         self->priv->modem = g_value_dup_object (value);
         break;
+    case PROP_HOTPLUGGED:
+        self->priv->hotplugged = g_value_get_boolean (value);
+        break;
+    case PROP_VIRTUAL:
+        self->priv->virtual = g_value_get_boolean (value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;
@@ -661,6 +783,12 @@ get_property (GObject *object,
     case PROP_MODEM:
         g_value_set_object (value, self->priv->modem);
         break;
+    case PROP_HOTPLUGGED:
+        g_value_set_boolean (value, self->priv->hotplugged);
+        break;
+    case PROP_VIRTUAL:
+        g_value_set_boolean (value, self->priv->virtual);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;
@@ -686,8 +814,9 @@ finalize (GObject *object)
 {
     MMDevice *self = MM_DEVICE (object);
 
-    g_free (self->priv->udev_device_path);
+    g_free (self->priv->path);
     g_strfreev (self->priv->drivers);
+    g_strfreev (self->priv->virtual_ports);
 
     G_OBJECT_CLASS (mm_device_parent_class)->finalize (object);
 }
@@ -705,6 +834,14 @@ mm_device_class_init (MMDeviceClass *klass)
     object_class->finalize = finalize;
     object_class->dispose = dispose;
 
+    properties[PROP_PATH] =
+        g_param_spec_string (MM_DEVICE_PATH,
+                             "Path",
+                             "Device path",
+                             NULL,
+                             G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
+    g_object_class_install_property (object_class, PROP_PATH, properties[PROP_PATH]);
+
     properties[PROP_UDEV_DEVICE] =
         g_param_spec_object (MM_DEVICE_UDEV_DEVICE,
                              "UDev Device",
@@ -729,6 +866,22 @@ mm_device_class_init (MMDeviceClass *klass)
                              G_PARAM_READWRITE);
     g_object_class_install_property (object_class, PROP_MODEM, properties[PROP_MODEM]);
 
+    properties[PROP_HOTPLUGGED] =
+        g_param_spec_boolean (MM_DEVICE_HOTPLUGGED,
+                              "Hotplugged",
+                              "Whether the modem was hotplugged",
+                              FALSE,
+                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
+    g_object_class_install_property (object_class, PROP_HOTPLUGGED, properties[PROP_HOTPLUGGED]);
+
+    properties[PROP_VIRTUAL] =
+        g_param_spec_boolean (MM_DEVICE_VIRTUAL,
+                              "Virtual",
+                              "Whether the device is virtual or real",
+                              FALSE,
+                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
+    g_object_class_install_property (object_class, PROP_VIRTUAL, properties[PROP_VIRTUAL]);
+
     signals[SIGNAL_PORT_GRABBED] =
         g_signal_new (MM_DEVICE_PORT_GRABBED,
                       G_OBJECT_CLASS_TYPE (object_class),
